use std::hash::Hash;

use crate::entity::{Entity, EntityColumn, EntityID};
use crate::model::Modelable;

// pub mod expr;
// pub mod condition;
pub mod builder;

/// Wraps an entity with its ID, for example as a query result.
///
/// The wrapped value is accessible via `Deref`, so this should be mostly
/// transparent.
#[derive(Debug)]
pub struct WithID<T: Entity> {
    wrap: T,
    id: <T as Entity>::ID,
}

impl<T: Entity> WithID<T> {
    pub fn new(wrap: T, id: <T as Entity>::ID) -> Self {
        Self { wrap, id }
    }

    fn wrap(what: T, raw_id: i64) -> Self {
        Self {
            wrap: what,
            id: <T as Entity>::ID::from_raw_id(raw_id),
        }
    }

    pub fn wrapped(self) -> T {
        self.wrap
    }
}

impl<T: Entity> WithID<T> {
    pub fn id(&self) -> <T as Entity>::ID {
        self.id
    }
}

impl<T: Entity> AsRef<T> for WithID<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> std::ops::Deref for WithID<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for WithID<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

type CacheIndex = (&'static str, std::any::TypeId, u64);

/// The query interface for a database.
///
/// As the query interface provides some level of caching, try to strive for as much sharing as
/// possible. Passing around `QueryInterface` references instead of `DB` references is a good way
/// to achieve this. However, `QueryInterface` is explicitly `!Send`, so you may need to use
/// something like a `DBPool`.
pub struct QueryInterface<'l> {
    db: &'l crate::DB,

    cache: std::sync::Mutex<std::collections::HashMap<CacheIndex, sqlite::Statement<'l>>>,

    // use a phantom non-Send-able type to implement !Send for QueryInterface
    prevent_send: std::marker::PhantomData<*mut ()>,
}

const NO_HASH: u64 = 0;

impl<'l> QueryInterface<'l> {
    pub fn new(db: &'l crate::DB) -> Self {
        Self {
            db,
            cache: std::sync::Mutex::new(std::collections::HashMap::new()),
            prevent_send: std::marker::PhantomData,
        }
    }

    /// Helper function to process an expected single result
    /// Note that this errors out if there is more than a single result
    fn expect_one_result<T>(
        &self,
        stmt: &mut sqlite::Statement,
        with_result: &mut dyn FnMut(&mut sqlite::Statement) -> Option<T>,
    ) -> Option<T> {
        let state = stmt.next().ok()?;
        if state != sqlite::State::Row {
            return None;
        }

        let res = with_result(stmt);

        let state = stmt.next().ok()?;
        if state != sqlite::State::Done {
            return None;
        }

        res
    }

    /// Helper function to process an expected zero results
    /// Note that this errors out if there is any result
    fn expect_no_result(&self, stmt: &mut sqlite::Statement) -> Option<()> {
        let state = stmt.next().ok()?;
        if state != sqlite::State::Done {
            return None;
        }

        Some(())
    }
}

impl<'l> QueryInterface<'l> {
    fn cached_query<Return>(
        &self,
        context: &'static str,
        ty: std::any::TypeId,
        create: &dyn Fn() -> sqlite::Statement<'l>,
        with: &mut dyn FnMut(&mut sqlite::Statement<'l>) -> Return,
    ) -> Return {
        let mut cache = self.cache.lock().expect("Couldn't acquire cache?");
        let key = (context, ty, NO_HASH);
        let query = cache.entry(key).or_insert_with(create);

        query.reset().expect("Couldn't reset query");
        with(query)
    }

    fn cached_query_column<T: Entity, Return>(
        &self,
        context: &'static str,
        ty: std::any::TypeId,
        variant: &[&dyn EntityColumn<Entity = T>],
        create: &dyn Fn() -> sqlite::Statement<'l>,
        with: &mut dyn FnMut(&mut sqlite::Statement<'l>) -> Return,
    ) -> Return {
        use std::hash::Hasher;

        let mut hasher = std::collections::hash_map::DefaultHasher::new();
        for v in variant {
            v.column_typeid().hash(&mut hasher);
        }
        let hash = hasher.finish();

        let mut cache = self.cache.lock().expect("Couldn't acquire cache?");
        let key = (context, ty, hash);
        let query = cache.entry(key).or_insert_with(create);

        query.reset().expect("Couldn't reset query");
        with(query)
    }
}

// Legacy interface
impl<'l> QueryInterface<'l> {
    /// Search for an entity by a property
    pub fn get_one_by<C: EntityColumn, V: Modelable>(
        &self,
        col: C,
        val: V,
    ) -> Option<WithID<C::Entity>> {
        let table_name = <C::Entity>::table_name();
        let column_name = col.name();

        self.cached_query_column(
            "get_one_by",
            std::any::TypeId::of::<C::Entity>(),
            &[&col],
            &|| {
                self.db
                    .conn
                    .prepare(&format!(
                        "SELECT * FROM \"{}\" WHERE \"{}\" = ?",
                        table_name, column_name
                    ))
                    .expect("")
            },
            &mut |stmt| {
                val.bind_to(stmt, 1).ok()?;

                self.expect_one_result(stmt, &mut |stmt| {
                    let id: i64 = stmt.read(0).ok()?;
                    Some(WithID::wrap(
                        <<C as EntityColumn>::Entity>::build_from(stmt).ok()?,
                        id,
                    ))
                })
            },
        )
    }

    /// Search for an entity by multiple properties
    pub fn get_one_by_multi<T: Entity>(
        &self,
        c: &[&dyn EntityColumn<Entity = T>],
        val: &[&dyn crate::model::Modelable],
    ) -> Option<WithID<T>> {
        let table_name = T::table_name();

        assert_eq!(c.len(), val.len());

        self.cached_query_column(
            "get_one_by_multi",
            std::any::TypeId::of::<T>(),
            c,
            &|| {
                let query = format!(
                    "SELECT * FROM \"{}\" WHERE {}",
                    table_name,
                    c.iter()
                        .map(|col| format!("\"{}\" = ?", col.name()))
                        .collect::<Vec<_>>()
                        .join(" AND ")
                );
                self.db
                    .conn
                    .prepare(&query)
                    .expect(format!("Failed to prepare SQL query: {}", query).as_str())
            },
            &mut |stmt| {
                for index in 0..val.len() {
                    val[index].bind_to(stmt, index + 1).ok()?;
                }

                self.expect_one_result(stmt, &mut |stmt| {
                    let id: i64 = stmt.read(0).ok()?;
                    Some(WithID::wrap(T::build_from(stmt).ok()?, id))
                })
            },
        )
    }

    /// Delete entities by searching with a single property
    pub fn delete_by<C: EntityColumn, V: Modelable>(&self, c: C, val: V) -> Option<()> {
        let table_name = <C::Entity>::table_name();
        let column_name = c.name();

        self.cached_query_column(
            "delete_by",
            std::any::TypeId::of::<C::Entity>(),
            &[&c],
            &|| {
                let query = format!("DELETE FROM \"{}\" WHERE {} = ?", table_name, column_name);
                self.db
                    .conn
                    .prepare(&query)
                    .expect(format!("Failed to prepare SQL query: {}", query).as_str())
            },
            &mut |stmt| {
                val.bind_to(stmt, 1).ok()?;

                self.expect_no_result(stmt)
            },
        )
    }

    /// Delete entities by searching with a single property
    pub fn delete_by_id<I: EntityID<Entity = T>, T: Entity<ID = I>>(
        &self,
        id: <T as Entity>::ID,
    ) -> Option<()> {
        let table_name = <T as Entity>::table_name();

        self.cached_query(
            "delete_by_id",
            std::any::TypeId::of::<T>(),
            &|| {
                let query = format!("DELETE FROM \"{}\" WHERE id = ?", table_name);
                self.db
                    .conn
                    .prepare(&query)
                    .expect(format!("Failed to prepare SQL query: {}", query).as_str())
            },
            &mut |stmt| {
                id.bind_to(stmt, 1).ok()?;

                self.expect_no_result(stmt)
            },
        )
    }

    /// Delete entities by searching with multiple properties
    pub fn delete_by_multi<T: Entity>(
        &self,
        c: &[&dyn EntityColumn<Entity = T>],
        val: &[&dyn crate::model::Modelable],
    ) -> Option<()> {
        let table_name = <T>::table_name();

        assert_eq!(c.len(), val.len());

        self.cached_query_column(
            "delete_by_multi",
            std::any::TypeId::of::<T>(),
            c,
            &|| {
                let query = format!(
                    "DELETE FROM \"{}\" WHERE {}",
                    table_name,
                    c.iter()
                        .map(|col| format!("\"{}\" = ?", col.name()))
                        .collect::<Vec<_>>()
                        .join(" AND ")
                );
                self.db
                    .conn
                    .prepare(&query)
                    .expect(format!("Failed to prepare SQL query: {}", query).as_str())
            },
            &mut |stmt| {
                for index in 0..val.len() {
                    val[index].bind_to(stmt, index + 1).ok()?;
                }

                self.expect_no_result(stmt)
            },
        )
    }

    /// Search for an entity by ID
    pub fn get_one_by_id<I: EntityID<Entity = T>, T: Entity>(&self, id: I) -> Option<WithID<T>> {
        let table_name = <T as Entity>::table_name();

        self.cached_query(
            "get_one_by_id",
            std::any::TypeId::of::<T>(),
            &|| {
                self.db
                    .conn
                    .prepare(&format!("SELECT * FROM \"{}\" WHERE id = ?", table_name))
                    .expect("")
            },
            &mut |stmt| {
                id.bind_to(stmt, 1).ok()?;

                self.expect_one_result(stmt, &mut |stmt| {
                    let id: i64 = stmt.read(0).ok()?;
                    Some(WithID::wrap(T::build_from(stmt).ok()?, id))
                })
            },
        )
    }

    /// Search for all entities matching a property
    pub fn get_all_by<C: EntityColumn, V: Modelable>(
        &self,
        c: C,
        val: V,
    ) -> Option<Vec<WithID<C::Entity>>> {
        let table_name = <C::Entity>::table_name();
        let column_name = c.name();

        self.cached_query_column(
            "get_all_by",
            std::any::TypeId::of::<C::Entity>(),
            &[&c],
            &|| {
                self.db
                    .conn
                    .prepare(&format!(
                        "SELECT * FROM \"{}\" WHERE {} = ?",
                        table_name, column_name
                    ))
                    .expect("")
            },
            &mut |stmt| {
                val.bind_to(stmt, 1).ok()?;

                let mut res = Vec::new();
                loop {
                    let state = stmt.next().ok()?;
                    if state == sqlite::State::Done {
                        break;
                    }

                    let id: i64 = stmt.read(0).ok()?;
                    res.push(WithID::wrap(C::Entity::build_from(stmt).ok()?, id));
                }

                Some(res)
            },
        )
    }

    /// Add an entity to its table
    pub fn add<T: Entity + serde::Serialize>(&self, m: &T) -> Option<<T as Entity>::ID> {
        self.cached_query(
            "add",
            std::any::TypeId::of::<T>(),
            &|| {
                let placeholders = (0..(<T as Entity>::column_count() - 1))
                    .map(|_| "?".to_string())
                    .collect::<Vec<_>>()
                    .join(",");

                self.db
                    .conn
                    .prepare(&format!(
                        "INSERT INTO \"{}\" VALUES (NULL, {}) RETURNING \"id\"",
                        <T as Entity>::table_name(),
                        placeholders
                    ))
                    .expect("")
            },
            &mut |stmt| {
                crate::model::store::serialize_into(stmt, m).ok()?;

                let rowid = self.expect_one_result(stmt, &mut |stmt| stmt.read::<i64>(0).ok())?;

                Some(<T as Entity>::ID::from_raw_id(rowid))
            },
        )
    }
}

/*

qi.get().by(Table::Column, value).by(Table::OtherColumn, value).one()
qi.get().id(some_id).one()
qi.get().by(Table::Column, value).all()

 * */

pub struct BuildResult(String, bool);

pub trait Buildable {
    fn build() -> BuildResult;
}

pub trait Gettable<'sq, 'l, T: Entity> {
    type Builder: Buildable;
    fn qi(&self) -> &'sq QueryInterface<'l> { todo!() }
}

pub struct GetQuery<'sq, 'l, T: Entity> {
    qi: &'sq QueryInterface<'l>,
    _ghost: std::marker::PhantomData<(T, &'sq str)>,
}


impl<'sq, 'l, T: Entity> Gettable<'sq, 'l, T> for GetQuery<'sq, 'l, T> {
    type Builder = GetBuild<T>;
}

pub struct GetBuild<T: Entity> { _ghost: std::marker::PhantomData<T> }

impl<T: Entity> Buildable for GetBuild<T> {
    fn build() -> BuildResult {
        todo!()
    }
}

pub struct ByClause<'sq, 'l, T: Entity, Col: EntityColumn<Entity = T>, Wrap: Gettable<'sq, 'l, T>> {
    wrap: Wrap,
    col: &'sq Col,
    data: &'sq dyn Modelable,
    _ghost: std::marker::PhantomData<(&'sq str, &'l str, T, Col, Wrap)>,
}

pub struct ByBuild<Col: EntityColumn, Wrap: Buildable> {
    _ghost: std::marker::PhantomData<(Col, Wrap)>
}

impl<Col: EntityColumn, Wrap: Buildable> Buildable for ByBuild<Col, Wrap> {
    fn build() -> BuildResult {
        Wrap::build();
        todo!()
    }
}

impl<'sq, 'l, T: Entity> GetQuery<'sq, 'l, T> {
    pub fn by<NCol: EntityColumn<Entity = T>>(self, col: &'sq NCol, data: &'sq dyn Modelable) -> ByClause<'sq, 'l, T, NCol, Self> {
        ByClause { wrap: self, col, data, _ghost: std::marker::PhantomData }
    }
}

impl<'sq, 'l, T: Entity, Col: EntityColumn<Entity = T>, Wrap: Gettable<'sq, 'l, T>> Gettable<'sq, 'l, T> for ByClause<'sq, 'l, T, Col, Wrap> {
    type Builder = ByBuild<Col, Wrap::Builder>;

    /*fn build_query(&self) -> (String, bool) {
        let (parent, is_where) = self.wrap.build_query();
        (format!("{} {} {} = ?", parent, if is_where { "AND" } else { "WHERE" }, "<>"), true)
    }*/
}

impl<'sq, 'l, T: Entity, Col: EntityColumn<Entity = T>, Wrap: Gettable<'sq, 'l, T>> ByClause<'sq, 'l, T, Col, Wrap> {
    fn build_query(&self) {
        // <Self as Gettable<T>>::build_query(self);
        // <Self as Gettable<T>>::Builder;
        <<Self as Gettable<T>>::Builder>::build();
    }

    pub fn one(self) -> Option<T> {
        // let ty = std::any::TypeId::of::<ByClause<'static, 'static, T, Col, Wrap>>();
        todo!()
    }

    pub fn all(self) -> Vec<T> {
        todo!()
    }

    pub fn by<NCol: EntityColumn<Entity = T>>(self, col: &'sq NCol, data: &'sq dyn Modelable) -> ByClause<'sq, 'l, T, NCol, Self> {
        ByClause { wrap: self, col, data, _ghost: std::marker::PhantomData }
    }
}

impl<'l> QueryInterface<'l> {
    // pub fn get<'q, T: Entity>(&'q self) -> GetQuery<T> where 'q: 'l {
    pub fn get<'sq, T: Entity>(&'sq self) -> GetQuery<T> where 'sq: 'l {
        GetQuery::<'sq, 'l, _> { qi: self, _ghost: std::marker::PhantomData }
    }
}

#[cfg(test)]
mod get_test {
    use microrm_macros::Entity;
    use serde::{Serialize, Deserialize};

    use crate::QueryInterface;


    #[derive(Entity, Serialize, Deserialize)]
    #[microrm_internal]
    pub struct KVStore {
        key: String,
        value: String
    }

    fn simple_test<'sq, 'l>(qi: &'sq QueryInterface<'l>) {
        let g = qi.get::<'sq, KVStore>();
        // qi.get::<'l,'_,KVStore>();

        //let t = g.by(&KVStore::Key, &"test");
        //drop(t);
        
    }

    #[test]
    fn test_build_query() {
        let db = crate::DB::new_in_memory(crate::Schema::new()).unwrap();
        let qi = db.query_interface();

        simple_test(&qi);

        {
        
        }
    }
}
